/*
 * C implementation of a bencode decoder.
 * This is the format defined by BitTorrent:
 *  http://wiki.theory.org/BitTorrentSpecification#bencoding
 *
 * The only external requirements are a few [standard] function calls and
 * the long long type.  Any sane system should provide all of these things.
 *
 * See the bencode.h header file for usage information.
 *
 * This is released into the public domain.
 * Written by Mike Frysinger <vapier@gmail.com>.
 */

/*
 * This implementation isn't optimized at all as I wrote it to support
 * a bogus system.  I have no real interest in this format.  Feel free
 * to send me patches (so long as you don't copyright them and you release
 * your changes into the public domain as well).
 */

#include <stdlib.h> /* malloc() realloc() free() strtoll() */
#include <string.h> /* memset() */
#include <unistd.h>
#include <stdio.h>
#include <limits.h>
#include <errno.h>

#include "bencode.h"

static be_node *be_alloc(be_type type)
{
	be_node *ret = malloc(sizeof(*ret));
	if (ret) {
		memset(ret, 0x00, sizeof(*ret));
		ret->type = type;
	}
	return ret;
}

static int64_t _be_decode_int(const char **data)
{
	char buf[128];
	char *ptr = buf;
	char *endp = *data;
	int negative = 0;
	int64_t	ret = 0;
	
	if (*endp == '+') {
	    *ptr = *endp;
	    ptr++;
	    endp++;
	}
	if (*endp == '-') {
	    *ptr = *endp;
	    ptr++;
	    endp++;
	    negative = 1;
	}
	while ( isdigit(*endp)) {
	    *ptr = *endp;
	    ret = (ret * 10) + (int64_t)(*endp-(char)'0');
	    ptr++;
	    endp++;
	}
	*ptr = (char)0;
	*data = endp;
	return ret;
}

static size_t _be_encode_int(blob_t * b, int64_t val)
{
	char buf[128];
	size_t res = 0;
	
	if ( b != NULL) {
	    res = snprintf( buf, sizeof(buf), "i%" PRIi64 "e", val);
	    if ( res > 0) {
		blob_write( b, buf, res);
		return res;
	    }
	}
	return 0;
}

static char *_be_decode_str(const char **data, int64_t * slen)
{
	/* XXX: should check for overflow/negative ... */
	*slen = _be_decode_int(data);
	char *ret = NULL;
	if (**data == ':') {
	 	ret = malloc(*slen + 1);
		memcpy(ret, *data + 1, *slen);
		ret[*slen] = '\0';
		*data += *slen + 1;
	}
	return ret;
}

static size_t _be_encode_str(blob_t * b, const char * val, int64_t len)
{
	char buf[128];
	size_t res = 0;
	int64_t	slen = 0;
	
	if ((b != NULL) && (val != NULL)) {
	    slen = (len == 0) ? (int64_t)strlen(val) : len;
	    res = snprintf( buf, sizeof(buf), "%" PRIi64 ":", slen);
	    if ( res > 0) {
		blob_write( b, buf, res);
		blob_write( b, val, slen);
		return res+(size_t)slen;
	    }
	}
	return 0;
}

static be_node *_be_decode(const char **data)
{
	be_node *ret = NULL;
	int64_t	len;

	switch (**data) {
		/* lists */
		case 'l': {
			unsigned int i = 0;

			ret = be_alloc(BE_LIST);

			++(*data);
			while (**data != 'e') {
				ret->val.l = realloc(ret->val.l, (i + 2) * sizeof(*ret->val.l));
				ret->val.l[i] = _be_decode(data);
				++i;
			}
			++(*data);

			ret->val.l[i] = NULL;

			return ret;
		}

		/* dictionaries */
		case 'd': {
			unsigned int i = 0;

			ret = be_alloc(BE_DICT);

			++(*data);
			while (**data != 'e') {
				ret->val.d = realloc(ret->val.d, (i + 2) * sizeof(*ret->val.d));
				ret->val.d[i].key = _be_decode_str(data, &len);
				ret->val.d[i].val = _be_decode(data);
				++i;
			}
			++(*data);

			ret->val.d[i].val = NULL;

			return ret;
		}

		/* integers */
		case 'i': {
			ret = be_alloc(BE_INT);

			++(*data);
			ret->i = _be_decode_int(data);
			if (**data != 'e')
				return NULL;
			++(*data);

			return ret;
		}

		/* byte strings */
		case '0'...'9': {
			ret = be_alloc(BE_STR);

			ret->val.s = _be_decode_str(data, &ret->i);

			return ret;
		}

		/* invalid */
		default:
			break;
	}

	return ret;
}

be_node *be_decode(const char *data)
{
	return _be_decode(&data);
}

void be_free(be_node *node)
{
	switch (node->type) {
		case BE_STR:
			free(node->val.s);
			break;

		case BE_INT:
			break;

		case BE_LIST: {
			unsigned int i;
			for (i = 0; node->val.l[i]; ++i)
				be_free(node->val.l[i]);
			free(node->val.l);
			break;
		}

		case BE_DICT: {
			unsigned int i;
			for (i = 0; node->val.d[i].val; ++i) {
				free(node->val.d[i].key);
				be_free(node->val.d[i].val);
			}
			free(node->val.d);
			break;
		}
	}
	free(node);
}

size_t be_encode(blob_t * b, be_node *node)
{
	size_t i = 0, k = 0;
	size_t	j;

    if (b == NULL)
	return i;
	
	switch (node->type) {
		case BE_STR:
			i = _be_encode_str(b, node->val.s, node->i);
			break;

		case BE_INT:
			i = _be_encode_int(b, node->i);
			break;

		case BE_LIST:
			blob_write( b, "l", 1);
			i++;
			for (j = 0; node->val.l[j]; ++j) {
				k = be_encode(b, node->val.l[j]);
				if ( k > 0)
				    i += k;
				else
				    return 0;
			}
			blob_write(b, "e", 1);
			i++;
			break;

		case BE_DICT:
			blob_write( b, "d", 1);
			i++;
			for (j = 0; node->val.d[j].val; ++j) {
				k = _be_encode_str(b, node->val.d[j].key, 0);
				if ( k > 0)
				    i += k;
				else
				    return 0;
				k = be_encode(b, node->val.d[j].val);
				if ( k > 0)
				    i += k;
				else
				    return 0;
			}

			blob_write( b, "e", 1);
			i++;
			break;
	}
    if ( i < 0)
	i = 0;
    return i;
}

#ifdef BE_DEBUG
#include <stdio.h>
#include <stdint.h>

static void _be_dump_indent(ssize_t indent)
{
	while (indent-- > 0)
		printf("    ");
}
static void _be_dump(be_node *node, ssize_t indent)
{
	size_t i;

	_be_dump_indent(indent);
	indent = abs(indent);

	switch (node->type) {
		case BE_STR:
			printf("str = %" PRIi64 "\n", node->i);
			break;

		case BE_INT:
			printf("int = %" PRIi64 "\n", node->i);
			break;

		case BE_LIST:
			puts("list [");

			for (i = 0; node->val.l[i]; ++i)
				_be_dump(node->val.l[i], indent + 1);

			_be_dump_indent(indent);
			puts("]");
			break;

		case BE_DICT:
			puts("dict {");

			for (i = 0; node->val.d[i].val; ++i) {
				_be_dump_indent(indent + 1);
				printf("%s => ", node->val.d[i].key);
				_be_dump(node->val.d[i].val, -(indent + 1));
			}

			_be_dump_indent(indent);
			puts("}");
			break;
	}
}
void be_dump(be_node *node)
{
	_be_dump(node, 0);
}
#endif

static char ann_list[] = "announce-list";

be_node * _be_search_ann_list( be_node * root) {
    be_node * node = NULL;
    size_t	i;
    
    if ( root != NULL) {
	switch ( root->type) {
	    case BE_DICT:
		for (i = 0; root->val.d[i].val; i++) {
		    if (strncmp(root->val.d[i].key, ann_list, 13) == 0)
			return root->val.d[i].val;
		    else {
			node = _be_search_ann_list( root->val.d[i].val);
			if ( node != NULL)
			    break;
		    }
		}
	    break;
	    case BE_LIST:
		for (i = 0; root->val.l[i]; i++) {
		    node = _be_search_ann_list( root->val.l[i]);
		    if ( node != NULL)
			break;
		}
	    break;
	    default:
	    break;
	}
    }
    return node;
}

static be_dict * _be_create_ann_list( void) {
    be_dict * node = NULL;

    if ((node = malloc(sizeof(be_dict))) != NULL) {
	if ((node->key = strdup(ann_list)) == NULL) {
	    free(node);
	    return NULL;
	}
	if ((node->val = be_alloc( BE_LIST)) == NULL) {
	    free(node->key);
	    free(node);
	    return NULL;
	}
    }
    return node;
}

static be_node * _be_create_retrecker( const char * str) {
    be_node * node = NULL;
    
    if ((node = be_alloc( BE_LIST)) != NULL) {
	if ((node->val.l = realloc(node->val.l, 2 * sizeof(*node->val.l))) == NULL) {
	    be_free(node);
	    return NULL;
	}
	node->val.l[1] = NULL;
	if ((node->val.l[0] = be_alloc( BE_STR)) != NULL)
	    node->val.l[0]->val.s = strdup( str);
	else {
	    be_free(node);
	    return NULL;
	}
    }
    return node;
}

static size_t	_be_size( be_node * node) {
    size_t	i = 0;
    
    if ( node != NULL) {
	switch ( node->type) {
	    case BE_DICT:
		for (; node->val.d[i].val; ++i) ;
	    break;
	    case BE_LIST:
		for (; node->val.l[i]; ++i) ;
	    break;
	    default:
	    break;
	}
    }
    return i;
}

static be_node * _be_list_search_str( be_node * node, const char * str) {
    size_t	i, slen;
    
    if ((str != NULL) && (node != NULL) && (node->type == BE_LIST)) {
	slen = strlen(str);
	for (i = 0; node->val.l[i]; ++i) {
	    if ( node->val.l[i]->type == BE_STR) {
		if ( strncmp(node->val.l[i]->val.s, str, slen) == 0)
		    return node->val.l[i];
	    }
	}
    }
    return NULL;
}

static void _be_remove_list_str( be_node * node, const char * str) {
    size_t	i, j, slen;
    
    if ((str != NULL) && (node != NULL) && (node->type == BE_LIST)) {
	slen = _be_size( node);
	for (i = 0; node->val.l[i]; ++i) {
	    if ( _be_list_search_str( node->val.l[i], str) != NULL) {
		be_free(node->val.l[i]);
		for ( j = i; j < slen; j++)
		    node->val.l[j] = node->val.l[j+1];
	    }
	}
    }
}

void be_process( be_node * root, const char * str, int remove_local) {
    int		do_remove = remove_local;
    be_node	* node, * ret;
    be_dict	* dict;
    size_t	sz, k;

    if ((root == NULL) || (root->type != BE_DICT))
	return;    
    node = _be_search_ann_list( root);
    if ((node == NULL) && (str != NULL)) {
	do_remove = 0;
	sz = _be_size(root);
	if (sz < 1)
	    return;
	root->val.d = realloc(root->val.d, (sz + 2) * sizeof(*root->val.d));
	for (k=1; k < sz; k++)
	    root->val.d[k+1] = root->val.d[k];
	if ((dict = _be_create_ann_list()) == NULL)
	    return;
	root->val.d[1].key = dict->key;
	root->val.d[1].val = dict->val;
	free( dict);
	node = _be_search_ann_list( root);
    }
    if (node == NULL)
	return;
    if ( do_remove)
	_be_remove_list_str( node, "http://retracker.local");
    if ( str != NULL) {
	if ((sz = _be_size( node)) == 0)
	    return;
	if ((ret = _be_create_retrecker(str)) == NULL)
	    return;
	node->val.l = realloc(node->val.l, (sz + 2) * sizeof(*node->val.l));
	node->val.l[sz] = ret;
	node->val.l[sz+1] = NULL;
    }
}
